/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                                                                                                         */
/*   Copyright (C) 2014 Thorsten Koch, Daniel Rehfeldt All Rights Reserved                                                                */
/*                                                                                                                                                         */
/*   Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"),    */
/*   to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense,    */
/*   and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:            */
/*                                                                                                                                                         */
/*   The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.                        */
/*                                                                                                                                                         */
/*   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,   */
/*   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER    */
/*   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER           */
/*   DEALINGS IN THE SOFTWARE.                                                                                                                             */
/*                                                                                                                                                         */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <stdint.h>

#if defined(_WIN32) || defined(_WIN64)
#include <windows.h>
//#pragma comment(linker, "/STACK:10000000") /* large stack for graph_trail */
#define R_OK 4  /* checks for read only */
#include <io.h>
#define access _access
#else
#include <sys/times.h>
#include <sys/time.h>
#include <unistd.h>
#endif
#include <time.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>         /* errno */
#include <assert.h>
#include <stdarg.h>        /* message: va_list etc */
#include <float.h>
/*#include <unistd.h>*/        /* R_OK  */


#define MINIMUM_TIME_PER_INSTANCE       5.0        /* minimum total time we must run for each instance */
#define MST_COMPUTATIONS                50         /* number of roots from which to start a MST computation (per instance) */
#define VERSION_NUMBER                  20140902

#define EPSILON 1e-04
#define EPS_ZERO  1e-02

/*
 * Maths
 */
//#define EPSILON                  1e-02
//#define EPS_ZERO                (DBL_EPSILON * 1e4)
#define LT(a, b)              (((a) - (b))     < -EPS_ZERO)
#define GT(a, b)              (((a) - (b))     >  EPS_ZERO)
#define EQ(a, b)   (fabs((a) - (b)) <= EPS_ZERO)
#define NE(a, b)   (fabs((a) - (b)) >  EPS_ZERO)

/*
 * Graph
 */
#define EAT_FREE     -1
#define EAT_LAST     -2
#define EAT_HIDE     -3

#define CONNECT       0
#define UNKNOWN     (-1)
#define FARAWAY       1e15

#define NO_CHANGE    -10

#define Is_term(a)   ((a) >= 0)
#define Edge_anti(a) ((((a) % 2) > 0) ? (a) - 1 : (a) + 1)



/*
 * Graph reader
 */
#define STP_MAGIC       0x33d32945
#define VERSION_MAJOR   1
#define VERSION_MINOR   0

#define TEST_INSTANCES "stp_test_instances.txt"

#define MSG_FATAL       0
#define MSG_ERROR       1
#define MSG_WARN        2
#define MSG_INFO        3
#define MSG_DEBUG       4

#define VERBOSE         0 //MSG_INFO

#ifndef TRUE
#define TRUE            1
#endif
#ifndef FALSE
#define FALSE           0
#endif

#ifndef SUCCESS
#define SUCCESS         0
#endif
#ifndef FAILURE
#define FAILURE       (-1)
#endif

#if defined(MSDOS) || defined(WIN32)
#define  DIRSEP "\\"
#else
#define  DIRSEP "/"
#endif

/* Try to get the maximum length of a path.
 *
 * WARNING: if a path found or build during the scanning process is
 *          longer than defined below the programm will probably
 *          crash, because scanf() will overwrite some memory.
 */
#if defined(PATH_MAX)                     /* Found this on SCO UNIX */
#define MAX_PATH_LEN PATH_MAX
#elif defined(_MAX_PATH)                  /* Watcom C on MSDOS */
#define MAX_PATH_LEN _MAX_PATH
#elif defined(_POSIX_PATH_MAX)            /* Maybe POSIX */
#define MAX_PATH_LEN _POSIX_PATH_MAX
#else
#define MAX_PATH_LEN 1024
#endif

/* Extension separators
 */
#define EXTSEP             '.'

#define MAX_LINE_LEN             1024
#define MAX_KEYWORD_LEN          64
#define MAX_STRING_LEN           256
#define MAX_ARGUMENTS            8



#define STP_UNDIRECTED               0
#define STP_DIRECTED                 1
#define STP_PRIZE_COLLECTING         2
#define STP_ROOTED_PRIZE_COLLECTING  3
#define STP_NODE_WEIGHTS             4
#define STP_DEG_CONS                 5
#define STP_REVENUES_BUDGET_HOPCONS  6
#define STP_GRID                     7
#define STP_OBSTACLES_GRID           8
#define STP_MAX_NODE_WEIGHT          9
#define STP_HOP_CONS                 10


#define KEY_SECTION              0001
#define KEY_EOF                  9999
#define KEY_END                  9998

#define KEY_COMMENT_NAME         1001
#define KEY_COMMENT_DATE         1002
#define KEY_COMMENT_CREATOR      1003
#define KEY_COMMENT_PROBLEM      1004
#define KEY_COMMENT_REMARK       1005

#define KEY_GRAPH_NODES          2001
#define KEY_GRAPH_EDGES          2002
#define KEY_GRAPH_E              2003
#define KEY_GRAPH_A              2004
#define KEY_GRAPH_OBSTACLES      2005
#define KEY_GRAPH_HOPLIMIT       2006
#define KEY_GRAPH_ROOT           2007
#define KEY_GRAPH_BUDGET         2008

#define KEY_TERMINALS_END        3001
#define KEY_TERMINALS_TERMINALS  3002
#define KEY_TERMINALS_T          3003
#define KEY_TERMINALS_TP         3004
#define KEY_TERMINALS_ROOT       3005
#define KEY_TERMINALS_ROOTP      3006

#define KEY_COORDINATES_DD       4001
#define KEY_COORDINATES_DDD      4002
#define KEY_COORDINATES_DDDD     4003
#define KEY_COORDINATES_DDDDD    4004
#define KEY_COORDINATES_DDDDDD   4005
#define KEY_COORDINATES_DDDDDDD  4006
#define KEY_COORDINATES_DDDDDDDD 4007

#define KEY_COORDINATES_END      4011
#define KEY_COORDINATES_GRID     4012

#define KEY_SOLUTION_VALUE       4021
#define KEY_SOLUTION_DATE        4022
#define KEY_SOLUTION_TIME        4023
#define KEY_SOLUTION_STEINER     4024
#define KEY_SOLUTION_S           4025

#define KEY_PRESOLVE_DATE        5001
#define KEY_PRESOLVE_FIXED       5002
#define KEY_PRESOLVE_LOWER       5003
#define KEY_PRESOLVE_UPPER       5004
#define KEY_PRESOLVE_TIME        5005
#define KEY_PRESOLVE_EA          5006
#define KEY_PRESOLVE_EC          5007
#define KEY_PRESOLVE_ED          5008
#define KEY_PRESOLVE_ES          5009


#define KEY_NODEWEIGHTS_NW       6000

#define KEY_PRIZECOLL_MD         7000

#define KEY_MAXDEGS_MD           8000

#define KEY_OBSTACLES_RR         9000
#define KEY_OBSTACLES_END        9001

#define KEY_PVERTICES_NPV        10001
#define KEY_PVERTICES_PV         10002

#define FLAG_OPTIONAL            1
#define FLAG_REQUIRED            2

#define SECTION_MISSING          0
#define SECTION_EXISTEND         1



/*
 * Graph structs
 */

typedef struct
{
   /* Knots
    */
   double* knotcost;   /* Array [0..knots-1] of knot costs    */
   int     budget;
   int     npv;    /* number of profitable vertices */
   int     ksize;  /* Count of allocated knot slots               */
   int     knots;  /* Count of knots in graph                     */
   int     terms;  /* Count of terminals                          */
   int     source; /* The root                                    */
   int*    term;   /* Array [0..knots-1] of networknumber for     */
                   /* knot [i], -1 if [i] is never a terminal     */
   int*    mark;   /* Array [0..knots-1], normaly TRUE or FALSE   */
                   /* to mark knots for inclusion in the shortest */
                   /* path / minimum spanning tree routine        */
   int*    grad;   /* Array [0..knots-1] with degree of knot [i]  */
   int*    inpbeg; /* Array [0..knots-1] with starting slot index */
                   /* for the ieat array, -1 if not used          */
   int*    outbeg; /* Array [0..knots-1] with starting slot index */
                   /* for the oeat array, -1 if not used          */

   int*    maxdeg; /* Array [0..knots-1] containing the maximal
                      degrees of all nodes (only used for Degree-
                      Constraint STPs)                            */
   /* Edges
    */
   int     hoplimit;
   int     esize;  /* Count of allocated edge slots               */
   int     edges;  /* Count of edges in the graph                 */
   double* cost;   /* Array [0..edges-1] of positiv edge costs    */
   int*    tail;   /* Array [0..edges-1] of knot-number of tail   */
                   /* of edge [i]                                 */
   int*    head;   /* Array [0..edges-1] of knot-number of head   */
                   /* of edge [i]                                 */
   /* Knots/Edges
    */
   int*    ieat;   /* Array [0..edges-1], incomming edge          */
                   /* allocation table                            */
   int*    oeat;   /* Array [0..edges-1], outgoing edge           */
                   /* allocation table                            */

   /* data for math and mst computation
    */
   int* path_heap;
   int* path_state;

   /* type of Steiner Tree Problem  */
   int stp_type;
} GRAPH;

typedef struct presolve_info
{
   double fixed;
   double upper;
   double lower;
   int    time;
} PRESOL;

/* ONE segment of a path
 */
typedef struct mst_unit
{
   double       dist;         /* Distance to the end of the path             */
   signed int   edge;         /* First edge to go                            */
} MST;



/*
 * Graph reader structs
 */

struct key
{
   const char* keyword;
   int         sw_code;
   const char* format;
};

static const struct key keyword_table[] =
   {
      /*
       * *** The keywords MUST be sorted alphabetically ! ***
       */
      {  ".eof",                     KEY_EOF,                    NULL        },
      {  ".section",                 KEY_SECTION,                NULL        },

      {  "comment.creator",          KEY_COMMENT_CREATOR,        "s"         },
      {  "comment.date",             KEY_COMMENT_DATE,           "s"         },
      {  "comment.end",              KEY_END,                    NULL        },
      {  "comment.name",             KEY_COMMENT_NAME,           "s"         },
      {  "comment.problem",          KEY_COMMENT_PROBLEM,        "s"         },
      {  "comment.remark",           KEY_COMMENT_REMARK,         "s"         },

      {  "coordinates.dd",           KEY_COORDINATES_DD,         "nnn"       },
      {  "coordinates.ddd",          KEY_COORDINATES_DDD,        "nnnn"      },
      {  "coordinates.dddd",         KEY_COORDINATES_DDDD,       "nnnnn"     },
      {  "coordinates.ddddd",        KEY_COORDINATES_DDDDD,      "nnnnnn"    },
      {  "coordinates.dddddd",       KEY_COORDINATES_DDDDDD,     "nnnnnnn"   },
      {  "coordinates.ddddddd",      KEY_COORDINATES_DDDDDDD,    "nnnnnnnn"  },
      {  "coordinates.dddddddd",     KEY_COORDINATES_DDDDDDDD,   "nnnnnnnnn" },
      {  "coordinates.end",          KEY_COORDINATES_END,        NULL        },
      {  "coordinates.grid",         KEY_COORDINATES_GRID,       NULL        },

      {  "graph.a",                  KEY_GRAPH_A,                "nnn"       },
      {  "graph.budget",             KEY_GRAPH_BUDGET,           "n"         },
      {  "graph.e",                  KEY_GRAPH_E,                "nnn"       },
      {  "graph.edges",              KEY_GRAPH_EDGES,            "n"         },
      {  "graph.end",                KEY_END,                    NULL        },
      {  "graph.hoplimit",           KEY_GRAPH_HOPLIMIT,         "n"         },
      {  "graph.nodes",              KEY_GRAPH_NODES,            "n"         },
      {  "graph.obstacles",          KEY_GRAPH_OBSTACLES,        "n"         },
      {  "graph.root",               KEY_GRAPH_ROOT,             "n"         },

      {  "maximumdegrees.end",       KEY_END,                    NULL        },
      {  "maximumdegrees.md",        KEY_MAXDEGS_MD,             "n"         },

      {  "nodeweights.end",          KEY_END,                    NULL        },
      {  "nodeweights.nw",           KEY_NODEWEIGHTS_NW,         "n"         },

      {  "obstacles.end",            KEY_OBSTACLES_END,          NULL        },
      {  "obstacles.rr",             KEY_OBSTACLES_RR,           "nnnn"      },

      {  "presolve.date",            KEY_PRESOLVE_DATE,          "s"         },
      {  "presolve.ea",              KEY_PRESOLVE_ED,            "nnnn"      },
      {  "presolve.ec",              KEY_PRESOLVE_EC,            "nnn"       },
      {  "presolve.ed",              KEY_PRESOLVE_ED,            "nnn"       },
      {  "presolve.end",             KEY_END,                    NULL        },
      {  "presolve.es",              KEY_PRESOLVE_ES,            "nn"        },
      {  "presolve.fixed",           KEY_PRESOLVE_FIXED,         "n"         },
      {  "presolve.lower",           KEY_PRESOLVE_LOWER,         "n"         },
      {  "presolve.orgnodes",        KEY_EOF,                    "n"         },
      {  "presolve.time",            KEY_PRESOLVE_TIME,          "n"         },
      {  "presolve.upper",           KEY_PRESOLVE_UPPER,         "n"         },

      {  "profitablevertices.end",             KEY_END,                    NULL        },
      {  "profitablevertices.profitablevertices",      KEY_PVERTICES_NPV,         "n"      },
      {  "profitablevertices.pv",      KEY_PVERTICES_PV,         "nn"         },

      {  "solution.date",            KEY_SOLUTION_DATE,          "s"         },
      {  "solution.end",             KEY_END,                    NULL        },
      {  "solution.s",               KEY_SOLUTION_S,             "n"         },
      {  "solution.steiner",         KEY_SOLUTION_STEINER,       "n"         },
      {  "solution.time",            KEY_SOLUTION_TIME,          "n"         },
      {  "solution.value",           KEY_SOLUTION_VALUE,         "n"         },

      {  "terminals.end",            KEY_TERMINALS_END,          NULL        },
      {  "terminals.root",           KEY_TERMINALS_ROOT,         "n"         },
      {  "terminals.rootp",          KEY_TERMINALS_ROOTP,        "n"         },
      {  "terminals.t",              KEY_TERMINALS_T,            "n"         },
      {  "terminals.terminals",      KEY_TERMINALS_TERMINALS,    "n"         },
      {  "terminals.tp",             KEY_TERMINALS_TP,           "nn"        },
   };

struct section
{
   const char* name;
   const char* extension;
   const int   flag;
   int         mark;
};


/* Extension NULL = no separate file possible !
 */
static struct section section_table[] =
   {
      { "",            "stp", FLAG_REQUIRED, SECTION_EXISTEND },

      /*
       * *** The section names MUST be sorted alphabetically ! ***
       */
      { "comment",     NULL,  FLAG_REQUIRED, SECTION_MISSING },
      { "coordinates", "crd", FLAG_OPTIONAL, SECTION_MISSING },
      { "graph",       "grp", FLAG_REQUIRED, SECTION_MISSING },
      { "maximumdegrees", "mdg", FLAG_OPTIONAL, SECTION_MISSING },
      { "nodeweights", "nwg", FLAG_OPTIONAL, SECTION_MISSING },
      { "obstacles",   "obs", FLAG_OPTIONAL, SECTION_MISSING },
      { "presolve",    "prs", FLAG_OPTIONAL, SECTION_MISSING },
      { "profitablevertices",    "pro", FLAG_OPTIONAL, SECTION_MISSING },
      { "solution",    "slt", FLAG_OPTIONAL, SECTION_MISSING },
      { "terminals",   "trm", FLAG_OPTIONAL, SECTION_MISSING },
   };

typedef struct current_file
{
   char   filename[MAX_PATH_LEN];
   int    line;
   FILE*  fp;
   struct section* section;
} CURF;

typedef union parameter
{
   double n;  /* Could be long long */
   char   s[MAX_STRING_LEN];
} PARA;



/*
 * Graph methods
 */

/** initialize a graph */
static GRAPH* graph_init(
   int ksize,
   int esize,
   int a,
   int  b
   )
{
   GRAPH* p;
   int    i;

   assert(ksize > 0);
   assert(esize > 0);

   p = (GRAPH*)malloc(sizeof(*p));

   assert(p != NULL);

   p->ksize  = ksize;
   p->knots  = 0;
   p->terms  = 0;
   p->source = -1;
   p->hoplimit = -1;
   p->term   = (int*)malloc((size_t)ksize * sizeof(int));
   p->mark   = (int*)malloc((size_t)ksize * sizeof(int));
   p->grad   = (int*)malloc((size_t)ksize * sizeof(int));
   p->inpbeg = (int*)malloc((size_t)ksize * sizeof(int));
   p->outbeg = (int*)malloc((size_t)ksize * sizeof(int));
   p->knotcost = NULL;
   p->maxdeg = NULL;

   p->esize = esize;
   p->edges = 0;
   p->cost  = (double*)malloc((size_t)esize * sizeof(double));
   p->tail  = (int*)malloc((size_t)esize * sizeof(int));
   p->head  = (int*)malloc((size_t)esize * sizeof(int));

   p->ieat  = (int*)malloc((size_t)esize * sizeof(int));
   p->oeat  = (int*)malloc((size_t)esize * sizeof(int));

   p->path_heap = NULL;
   p->path_state = NULL;

   assert(p->term   != NULL);
   assert(p->mark   != NULL);
   assert(p->grad   != NULL);
   assert(p->inpbeg != NULL);
   assert(p->outbeg != NULL);
   assert(p->cost   != NULL);
   assert(p->tail   != NULL);
   assert(p->head   != NULL);
   assert(p->ieat   != NULL);
   assert(p->oeat   != NULL);

   return(p);
}

/** free the graph */
static void graph_free(
   GRAPH* p)
{
   assert(p != NULL);

   free(p->term);
   free(p->mark);
   free(p->grad);
   free(p->inpbeg);
   free(p->outbeg);
   free(p->cost);
   free(p->tail);
   free(p->head);
   free(p->ieat);
   free(p->oeat);
   if( p->knotcost != NULL );
   free(p->knotcost);
   if( p->maxdeg != NULL );
   free(p->maxdeg);
   free(p);
}

/** add a node */
static void graph_knot_add(
   GRAPH* p,
   int    term,
   int a,
   int b)
{
   assert(p        != NULL);
   assert(p->ksize >  p->knots);

   p->term  [p->knots] = term;
   p->mark  [p->knots] = TRUE;
   p->grad  [p->knots] = 0;
   p->inpbeg[p->knots] = EAT_LAST;
   p->outbeg[p->knots] = EAT_LAST;

   if (Is_term(term))
   {
      p->terms++;
   }
   p->knots++;
}

/** change the terminal property of the node */
static void graph_knot_chg(
   GRAPH* p,
   int    knot,
   int    term,
   int    a,
   int    b)
{
   assert(p      != NULL);
   assert(knot   >= 0);
   assert(knot   < p->knots);

   if (term != NO_CHANGE)
   {
      if (term != p->term[knot])
      {
         if (Is_term(p->term[knot]))
         {
            p->terms--;
         }
         p->term[knot] = term;

         if (Is_term(p->term[knot]))
         {
            p->terms++;
         }
      }
   }
}

inline static void edge_remove(
   GRAPH* p,
   int    e)
{
   int    i;
   int    head;
   int    tail;

   assert(p          != NULL);
   assert(e          >= 0);
   assert(e          <  p->edges);

   head = p->head[e];
   tail = p->tail[e];

   if (p->inpbeg[head] == e)
      p->inpbeg[head] = p->ieat[e];
   else
   {
      for(i = p->inpbeg[head]; p->ieat[i] != e; i = p->ieat[i])
         assert(i >= 0);

      p->ieat[i] = p->ieat[e];
   }
   if (p->outbeg[tail] == e)
      p->outbeg[tail] = p->oeat[e];
   else
   {
      for(i = p->outbeg[tail]; p->oeat[i] != e; i = p->oeat[i])
         assert(i >= 0);

      p->oeat[i] = p->oeat[e];
   }
}



void graph_edge_del(
   GRAPH* p,
   int    e)
{
   assert(p          != NULL);
   assert(e          >= 0);
   assert(e          <  p->edges);

   /* Immer mit der ersten von beiden Kanten anfangen
    */
   e -= e % 2;

   assert(p->head[e] == p->tail[e + 1]);
   assert(p->tail[e] == p->head[e + 1]);

   p->grad[p->head[e]]--;
   p->grad[p->tail[e]]--;

   edge_remove(p, e);

   assert(p->ieat[e] != EAT_FREE);
   assert(p->ieat[e] != EAT_HIDE);
   assert(p->oeat[e] != EAT_FREE);
   assert(p->oeat[e] != EAT_HIDE);

   p->ieat[e] = EAT_FREE;
   p->oeat[e] = EAT_FREE;

   e++;

   edge_remove(p, e);

   assert(p->ieat[e] != EAT_FREE);
   assert(p->ieat[e] != EAT_HIDE);
   assert(p->oeat[e] != EAT_FREE);
   assert(p->oeat[e] != EAT_HIDE);

   p->ieat[e] = EAT_FREE;
   p->oeat[e] = EAT_FREE;
}

/** add an edge */
static void graph_edge_add(
   GRAPH* p,
   int    tail,
   int    head,
   double cost1,
   double cost2)
{
   int    e;

   assert(p      != NULL);
   assert(tail   >= 0);
   assert(tail   <  p->knots);
   assert(head   >= 0);
   assert(head   <  p->knots);

   assert(p->esize >= p->edges + 2);

   e = p->edges;

   p->grad[head]++;
   p->grad[tail]++;

   p->cost[e]      = cost1;
   p->tail[e]      = tail;
   p->head[e]      = head;
   p->ieat[e]      = p->inpbeg[head];
   p->oeat[e]      = p->outbeg[tail];
   p->inpbeg[head] = e;
   p->outbeg[tail] = e;

   e++;

   p->cost[e]      = cost2;
   p->tail[e]      = head;
   p->head[e]      = tail;
   p->ieat[e]      = p->inpbeg[tail];
   p->oeat[e]      = p->outbeg[head];
   p->inpbeg[tail] = e;
   p->outbeg[head] = e;

   p->edges += 2;
}

/** mark all nodes connected to node i */
static void graph_trail(
   const GRAPH* p,
   int          i)
{
   int   k;

   assert(p      != NULL);
   assert(i      >= 0);
   assert(i      <  p->knots);

   if (!p->mark[i])
   {
      p->mark[i] = TRUE;

      for(k = p->outbeg[i]; k != EAT_LAST; k = p->oeat[k])
         if (!p->mark[p->head[k]])
            graph_trail(p, p->head[k]);
   }
}

/** check whether the graph is valid */
static int graph_valid(
   const GRAPH* p)
{
   const char* error1  = "*** Graph Validation Error: Head invalid, Knot %d, Edge %d, Tail=%d, Head=%d\n";
   const char* error2  = "*** Graph Validation Error: Tail invalid, Knot %d, Edge %d, Tail=%d, Head=%d\n";
   const char* error3  = "*** Graph Validation Error: Source invalid, Layer %d, Source %d, Terminal %d\n";
   const char* error4  = "*** Graph Validation Error: FREE invalid, Edge %d/%d\n";
   const char* error5  = "*** Graph Validation Error: Anti invalid, Edge %d/%d, Tail=%d/%d, Head=%d/%d\n";
   const char* error6  = "*** Graph Validation Error: Knot %d with Grad 0 has Edges\n";
   const char* error7  = "*** Graph Validation Error: Knot %d not connected\n";
   const char* error8  = "*** Graph Validation Error: Wrong Terminal count, count is %d, should be %d\n";

   int    k;
   int    l = 0;
   int    e;
   int    terms;

   assert(p      != NULL);

   terms  = p->terms;

   for(k = 0; k < p->knots; k++)
   {
      if (Is_term(p->term[k]))
      {
         terms--;
      }
      for(e = p->inpbeg[k]; e != EAT_LAST; e = p->ieat[e])
         if (p->head[e] != k)
            break;

      if (e != EAT_LAST)
         return((void)fprintf(stderr, error1, k, e, p->tail[e], p->head[e]), FALSE);

      for(e = p->outbeg[k]; e != EAT_LAST; e = p->oeat[e])
         if (p->tail[e] != k)
            break;

      if (e != EAT_LAST)
         return((void)fprintf(stderr, error2, k, e, p->tail[e], p->head[e]), FALSE);
   }

   if (terms != 0)
      return((void)fprintf(stderr, error8, p->terms, p->terms - terms), FALSE);

   if ((p->source < 0)
      || (p->source >= p->knots)
      || (p->term[p->source] != 0))
      return((void)fprintf(stderr, error3,
            l, p->source, p->term[p->source]), FALSE);

   for(e = 0; e < p->edges; e += 2)
   {
      if ((p->ieat[e    ] == EAT_FREE) && (p->oeat[e    ] == EAT_FREE)
         && (p->ieat[e + 1] == EAT_FREE) && (p->oeat[e + 1] == EAT_FREE))
         continue;

      if ((p->ieat[e] == EAT_FREE) || (p->oeat[e] == EAT_FREE)
         || (p->ieat[e + 1] == EAT_FREE) || (p->oeat[e + 1] == EAT_FREE))
         return((void)fprintf(stderr, error4, e, e + 1), FALSE);

      if ((p->head[e] != p->tail[e + 1]) || (p->tail[e] != p->head[e + 1]))
         return((void)fprintf(stderr, error5,
               e, e + 1, p->head[e], p->tail[e + 1],
               p->tail[e], p->head[e + 1]), FALSE);
   }
   for(k = 0; k < p->knots; k++)
      p->mark[k] = FALSE;

   graph_trail(p, p->source);

   for(k = 0; k < p->knots; k++)
   {
      if ((p->grad[k] == 0)
         && ((p->inpbeg[k] != EAT_LAST) || (p->outbeg[k] != EAT_LAST)))
         return((void)fprintf(stderr, error6, k), FALSE);

      if (!p->mark[k] && (p->grad[k] > 0))
         return((void)fprintf(stderr, error7, k), FALSE);
   }
   return(TRUE);
}



/*
 * Graph reader methods
 */

/*---------------------------------------------------------------------------*/
/*--- Name     : String to Lower                                          ---*/
/*--- Function : Converts a string to lower case.                         ---*/
/*--- Arguments: Pointer to string.                                       ---*/
/*--- Returns  : The argument, but now pointing to a lower case string.   ---*/
/*---------------------------------------------------------------------------*/
static char* strlower(
   char* s)
{
   char* t;

   for(t = s; *s != '\0'; s++)
      *s = (char)tolower(*s);

   return(t);
}

/*---------------------------------------------------------------------------*/
/*--- Name     : Print Message                                            ---*/
/*--- Function : Prints a message on stderr.                              ---*/
/*--- Arguments: Type of message, Info about filename and line number,    ---*/
/*---            printf format string and parameters to be printed.       ---*/
/*--- Returns  : Nothing                                                  ---*/
/*---------------------------------------------------------------------------*/
static void message(
   unsigned int type,
   const CURF*  curf,
   const char*  msg,
   ...)
{

   va_list params;

   const char* intro[] = { "Fatal Error", "Error      ", "Warning    ",
                           "Info       ", "Debug      "
   };
   const char* header  = "*** %s File \"%s\" line %05d: ";

   assert(type           <  sizeof(intro) / sizeof(char*));
   assert(curf           != NULL);
   assert(curf->filename != NULL);
   assert(curf->line     >= 0);
   assert(msg            != NULL);

   va_start(params, msg);

   if (type <= VERBOSE)
   {
      (void)fprintf(stderr, header, intro[type], curf->filename, curf->line);
      (void)vfprintf(stderr, msg, params);
      (void)fputc('\n', stderr);
   }
   va_end(params);
}

/*---------------------------------------------------------------------------*/
/*--- Name     : Key Compare                                              ---*/
/*--- Function : Compares the key with an element of the list.            ---*/
/*--- Parameter: Pointer to key, pointer to element                       ---*/
/*--- Returns  : <0 : key<elem, =0 : key=elem, >0 : key>elem              ---*/
/*---------------------------------------------------------------------------*/
static int key_cmp(
   const void* key,
   const void* elem)
{
   assert(key                                != NULL);
   assert(elem                               != NULL);
   assert(((const struct key*)elem)->keyword != NULL);
#if 0
   (void)fprintf(stderr, "key [%s] elem [%s]\n",
      (const char*)key, ((const struct key*)elem)->keyword);
#endif
   return(strcmp((const char*)key, ((const struct key*)elem)->keyword));
}

/*---------------------------------------------------------------------------*/
/*--- Name     : Section Compare                                          ---*/
/*--- Function : Compares the key with an section name.                   ---*/
/*--- Parameter: Pointer to key, pointer to section                       ---*/
/*--- Returns  : <0 : key<sec, =0 : key=sec, >0 : key>sec                 ---*/
/*---------------------------------------------------------------------------*/
static int sec_cmp(
   const void* key,
   const void* section)
{
   assert(key                                    != NULL);
   assert(section                                != NULL);
   assert(((const struct section*)section)->name != NULL);

   return(strcmp((const char*)key, ((const struct section*)section)->name));
}

/*---------------------------------------------------------------------------*/
/*--- Name     : Get Arguments                                            ---*/
/*--- Function : Extract the arguments following a keyword.               ---*/
/*--- Parameter: Current file info, argument format string, input line,   ---*/
/*---            pointer to array of MAX_ARGUMENTS PARA's.                ---*/
/*--- Returns  : 0 for success and < 0 for failure.                       ---*/
/*---------------------------------------------------------------------------*/
static int get_arguments(
   const CURF* curf,
   const char* format,
   const char* s,
   PARA*       para)
{
   const char* err_missmatch_v = "Wrong Syntax";
   const char* msg_hello_ss    = "get_arguments(\"%s\", \"%s\")";

   int missmatch = FALSE;
   int i;
   int decimal_spaces;
   char is_negative;
   assert(format != NULL);
   assert(s      != NULL);
   assert(para   != NULL);

   message(MSG_DEBUG, curf, msg_hello_ss, format, s);

   /* We try until we run out of input or have enough arguments.
    */
   while((*s != '\0') && (*format != '\0') && !missmatch)
   {
      missmatch = TRUE;

      switch(*format)
      {
      case 'n' :  /* Numeric */
         /* Go to next digit.
          */
         while((*s != '\0') && !isdigit(*s) && (*s != '.') && (*s != '-') )
	 {
            s++;
	 }
         /* Someting left ?
          */
         if (*s != '\0')
         {
            assert(isdigit(*s) || (*s == '.') || (*s == '-'));

            /* Get it.
             */
            para->n = 0;
            decimal_spaces = -1;
	    is_negative = FALSE;
            while(isdigit(*s) || (*s == '.') || (*s == '-'))
            {
	       if( *s == '.' )
		  decimal_spaces = 0;
	       else if( *s == '-' )
		  is_negative = TRUE;
	       else if( decimal_spaces != -1 )
		  para->n = para->n + pow(10, -(++decimal_spaces)) * (*s - '0');
	       else
                  para->n = para->n * 10 + (*s - '0');
               s++;
            }
            if( is_negative )
               para->n = (-1) * para->n;
            missmatch = FALSE;
         }
         break;
      case 's' :  /* String */
         /* Go to the beginning of the string.
          */
         while((*s != '\0') && (*s != '\"'))
            s++;

         /* Someting left ?
          */
         if (*s != '\0')
         {
            assert(*s == '\"');

            /* Get the String.
             */
            i = 0;
            s++;

            while((*s != '\0') && (*s != '\"') && (i < MAX_STRING_LEN - 1))
               para->s[i++] = *s++;

            para->s[i] = '\0';
            missmatch  = FALSE;
         }
         break;
      case 'b' :  /* Bool */
      case 'd' :  /* Date */
      default  :
         /* CONSTCOND */
         assert(FALSE);
         break;
      }
      if (missmatch)
         message(MSG_ERROR, curf, err_missmatch_v);
      else
      {
         para++;
         format++;

      }
   }
   return((*format == '\0') ? SUCCESS : FAILURE);
}

/*---------------------------------------------------------------------------*/
/*--- Name     : Open File                                                ---*/
/*--- Function : Opens a file, processes the file header and secures it's ---*/
/*---            the right file.                                          ---*/
/*--- Arguments: Info about filename and wanted section (file type).      ---*/
/*--- Returns  : 0 for success with curf->fp and curf->line filled or     ---*/
/*---            or < 0 for failure.                                      ---*/
/*---------------------------------------------------------------------------*/
static int open_file(
   CURF* curf)
{
   const char* err_cantopen_s   = "%s.";
   const char* err_noheader_v   = "Wrong file header.";
   const char* err_nomagic_d    = "Wrong Magic-Number %d.";
   const char* err_wrongtype_ss = "Wrong file type. Found %s, expected %s.";
   const char* msg_version_dd   = "Format Version %d.%d.";

   char         type[4];
   unsigned int magic;
   int          version_major;
   int          version_minor;
   int result = FAILURE;

   assert(curf           != NULL);
   assert(curf->filename != NULL);
   assert(curf->section  != NULL);

   /* Prepare the result.
    */
   curf->line = 1;
   curf->fp   = NULL;

   /* Try to open the file...
    */
   if ((curf->fp = fopen(curf->filename, "r")) == NULL)
      message(MSG_FATAL, curf, err_cantopen_s, strerror(errno));

   /* Read Header...
    */
   else if (fscanf(curf->fp, "%8x %3s File, STP Format Version %2d.%2d \n",
         &magic, type, &version_major, &version_minor) != 4)
      message(MSG_FATAL, curf, err_noheader_v);

   /* Test Magic...
    */
   else if (magic != STP_MAGIC)
      message(MSG_FATAL, curf, err_nomagic_d, magic);

   /* Did we get the right type of file ?
    */
   else if (strcmp(strlower(type), curf->section->extension))
      message(MSG_FATAL, curf, err_wrongtype_ss, type, curf->section->extension);
   else
   {
      /* Succeeded. Just give a warning if the file has a different
       * version number than the reader and hope it will be ok.
       */
      if ((version_major != VERSION_MAJOR) || (version_minor != VERSION_MINOR))
         message(MSG_WARN, curf, msg_version_dd, version_minor, version_major);

      result = SUCCESS;
   }
   return(result);
}

/*---------------------------------------------------------------------------*/
/*--- Name     : Start a new Section                                      ---*/
/*--- Function : Starts a new section, maybe in another file.             ---*/
/*--- Parameter: Path for files, Basename, pointer to actual file info,   ---*/
/*---            Pointer to where to save actual file info, inputline.    ---*/
/*--- Returns  : 0 for success and < 0 for failure.                       ---*/
/*---------------------------------------------------------------------------*/
static int start_section(
   const char* pathname,
   const char* basename,
   CURF*       curf,
   CURF*       save,
   const char* s)
{
   const char* err_missing_v   = "Section name missing";
   const char* err_badsect_s   = "Unknown section name [%s]";
   const char* err_required_s  = "Can't access required file [%s]";
   const char* err_duplicate_s = "Duplicate Section [%s]";

   CURF            temp;
   char            inclname[MAX_PATH_LEN];
   char            sectname[MAX_KEYWORD_LEN];
   char            dummy   [MAX_KEYWORD_LEN];
   int             tokens;
   int             ret = FAILURE;

   assert(pathname != NULL);
   assert(basename != NULL);
   assert(curf     != NULL);
   assert(save     != NULL);
   assert(s        != NULL);

   sectname[0] = '\0';
   inclname[0] = '\0';

   /* Extract names
    */
   if ((tokens = sscanf(s, "%63s %63s %s", sectname, dummy, inclname)) < 1)
      message(MSG_FATAL, curf, err_missing_v);
   else
   {
      if( strcmp(strlower(sectname),"comments") == 0 )
         sectname[7] = '\0';

      /* Known section ?
       */
      temp.section = (struct section*)bsearch(strlower(sectname),
         &section_table[1],
         (sizeof(section_table) / sizeof(struct section)) - 1,
         sizeof(struct section), sec_cmp);

      if (temp.section == NULL)
         message(MSG_FATAL, curf, err_badsect_s, sectname);
      else
      {
         if (0 && temp.section->mark & SECTION_EXISTEND)
            message(MSG_FATAL, curf, err_duplicate_s, sectname);
         else
         {
            /* Is this section in a separate file ?
             */
            if (tokens == 1)
            {
               curf->section        = temp.section;
               curf->section->mark |= SECTION_EXISTEND;
               ret                  = SUCCESS;
            }
            else
            {
               (void)sprintf(temp.filename, "%s%s%c%s",
                  pathname,
                  (*inclname == '\0') ? basename : inclname,
                  EXTSEP,
                  temp.section->extension);

               if (access(temp.filename, R_OK))
               {
                  /* We can't access the include file.
                   * If the section is optional, we just ignore
                   * the whole thing, otherwise we have a problem.
                   */
                  if (temp.section->flag & FLAG_REQUIRED)
                     message(MSG_FATAL, curf, err_required_s, temp.filename);
                  else
                  {
                     temp.section = &section_table[0];
                     ret          = SUCCESS;
                  }
               }
               else
               {
                  if (!open_file(&temp))
                  {
                     *save                = *curf;
                     *curf                = temp;
                     curf->section->mark |= SECTION_EXISTEND;

                     ret                  = SUCCESS;
                  }
               }
            }
         }
      }
   }
   return(ret);
}

/*---------------------------------------------------------------------------*/
/*--- Name     : Steiner Tree Problem Load                                ---*/
/*--- Function : Reads a file in STP format and parses it.                ---*/
/*--- Parameter: Pointer to filename, Pointer to presolve struct          ---*/
/*--- Returns  : 0 for success, < 0 for failure                           ---*/
/*---------------------------------------------------------------------------*/
GRAPH* graph_load(
   const char*  file,
   PRESOL*      presol)
{
   const char*  err_unknown_s   = "Unknown keyword [%s]";
   const char*  err_include_v   = "Include in included file";
   const char*  err_missing_s   = "Required section %s missing";
   const char*  msg_newsect_s   = "Processing Section %s";
   const char*  msg_keyword_sd  = "Found Keyword \"%s\", code = %d";
   const char*  err_badedge_ddd = "Bad edge %d-%d (%d nodes)";
   const char*  err_badroot_dd  = "Bad root %d (%d nodes)";
   const char*  err_baddeg_dd   = "More degree constraints (%d) than nodes (%d)";
   const char*  msg_finish_dddd = "Knots: %d  Edges: %d  Terminals: %d  Source=%d\n";

   const char*  endofline = "#;\n\r";
   const char*  separator = " \t:=";

   static CURF  curf_null = { "", 0, NULL, NULL };

   GRAPH*       g       = NULL;
   CURF         curf;
   CURF         save;
   PARA         para    [MAX_ARGUMENTS];
   double       nodeweight;
   double*      prize = NULL;                  /* needed in prize collecting  */
   double*      maxnodeweights = NULL;
   char         buffer  [MAX_LINE_LEN];
   char         pathname[MAX_PATH_LEN];
   char         basename[MAX_PATH_LEN];
   char         keyword [MAX_KEYWORD_LEN];
   int          stop_input = FALSE;
   int          ret        = FAILURE;
   unsigned int crc        = 0;
   char*        s;
   char*        t;
   struct key*  p;
   double**     coordinates = NULL;
   int          i;
   int          head;
   int          root = UNKNOWN;
   int          tail;
   int          npv;
   int          pv;
   int          grid_dim = -1;
   int          terms = 0;
   int          nodes = 0;
   int          edges = 0;
   int          nwcount = 0;
   int          budget = UNKNOWN;
   int          degcount = 0;
   int          stp_type = -1;
   int          hoplimit = -1;
   int          termcount = 0;
   int          nobstacles = -1;
   int          scale_order;
   int          is_gridgraph = FALSE;
   int          has_coordinates = FALSE;
   int          obstacle_counter = 0;
   int**        scaled_coordinates = NULL;
   int**        obstacle_coords = NULL;
   assert(file != NULL);

   /* No section loaded so far.
    */
   for(i = 1; i < (int)(sizeof(section_table) / sizeof(section_table[0])); i++)
      section_table[i].mark = SECTION_MISSING;

   /* Get the names...
    */
   (void)strcpy(pathname, file);

   /* Did we get a path ?
    */
   if ((s = strrchr(pathname, DIRSEP[0])) == NULL)
   {
      /* No, no path.
       */
      (void)strcpy(basename, pathname);
      pathname[0] = '\0';
   }
   else
   {
      /* Yes, there is a path in front.
       */
      s++;
      (void)strcpy(basename, s);
      *s = '\0';
   }

   /* Strip of the extension
    */
   if ((s = strrchr(basename, EXTSEP)) != NULL)
      *s = '\0';

   /* Build filename
    */
   curf.section     = &section_table[0];
   save             = curf_null;

   (void)sprintf(curf.filename, "%s%s.%s",
      pathname,
      basename,
      curf.section->extension);

   /* Open the file...
    */
   if (!open_file(&curf))
   {
      /* We read while a file is open...
       */
      while((curf.fp != NULL) && !stop_input)
      {
         /* Read a line.
          */
         if ((s = fgets(buffer, sizeof(buffer), curf.fp)) == NULL)
         {
            /* No more lines available, so we can close the file.
             */
            (void)fclose(curf.fp);

            /* If we are in an included file, we come back to our
             * main file. Otherwise fp_save is NULL and than fp will
             * get NULL so we're finished.
             */
            curf = save;
            save = curf_null;

            continue;
         }
         /* Count line number
          */
         curf.line++;

         /* Find the start of the interesting part of an inputline.
          */
         while(isspace(*s))
            s++;

         /* Find the end of the interesting portion of an input line.
          * Either the start of a comment or the final NL or CR.
          * Since all lines but the last must have at least a NL
          * t is nearly never NULL.
          */
         if ((t = strpbrk(s, endofline)) != NULL)
            *t = '\0';

         /* Is there an interesting part left ?
          */
         if (*s == '\0')
            continue;

         /* Computation of Checksum
          */
         // crc = STPcrc32((unsigned char*)s, crc);

         /* Build a keyword of form "sectionname.keyword"
          */
         (void)strcpy(keyword, curf.section->name);

         i          = (int)strlen(keyword);
         keyword[i] = '.';

         for(i++;
             (i < MAX_KEYWORD_LEN - 1) && (isalpha(*s) || (*s == '_'));
             i++, s++)
            keyword[i] = (char)tolower(*s);

         keyword[i] = '\0';

         /* Skip junk following the keyword.
          */
         while((*s != '\0') && (strchr(separator, *s) != NULL))
            s++;

         if( strcmp(keyword,"comments") == 0 )
            keyword[i - 1] = '\0';

         /* Did we know the keyword ?
          */
         p = (struct key*)bsearch(keyword, keyword_table,
            sizeof(keyword_table) / sizeof(struct key),
            sizeof(struct key), key_cmp);

         if (p == NULL)
            message(MSG_ERROR, &curf, err_unknown_s, keyword);
         else
         {
	    char* format = (char*) p->format;
            assert(p != NULL);

            message(MSG_DEBUG, &curf, msg_keyword_sd, p->keyword, p->sw_code);

            /* Yes, so lets get the rest of the line if possible
             */
	    if( stp_type == STP_MAX_NODE_WEIGHT && p->format != NULL && (p->sw_code == KEY_TERMINALS_T || p->sw_code == KEY_GRAPH_E) )
	    {
               if( p->sw_code == KEY_TERMINALS_T)
                  format = (char*)"nn";
               else if( p->sw_code == KEY_GRAPH_E )
                  format = (char*)"nn";
	    }

            if ((p->format == NULL)
               || !get_arguments(&curf, (const char*) format, s, para))
            {
               /* Now, what should we do ?
                */
               switch(p->sw_code)
               {
               case KEY_SECTION : /* a new Section starts. */
                  if (save.fp == NULL)
                     stop_input = start_section(pathname, basename,
                        &curf, &save, s);
                  else
                  {
                     message(MSG_FATAL, &curf, err_include_v);
                     stop_input = TRUE;
                  }
                  if (!stop_input)
                     message(MSG_INFO, &curf, msg_newsect_s,
                        curf.section->name);

                  /* Reset CRC
                   */
                  crc = 0;
                  break;
               case KEY_END : /* END found. */
                  // message(MSG_INFO, &curf, "CRC [%X]", crc);

                  curf.section = &section_table[0];
                  break;
               case KEY_EOF : /* EOF found */
                  ret        = SUCCESS;
                  stop_input = TRUE;

                  /* Test if all required section were found.
                   */
                  for(i = 0; (unsigned)i < sizeof(section_table) / sizeof(struct section); i++)
                  {
                     if ((section_table[i].flag & FLAG_REQUIRED)
                        && !(section_table[i].mark & SECTION_EXISTEND))
                     {
                        message(MSG_FATAL, &curf, err_missing_s, section_table[i].name);
                        ret = FAILURE;
                     }
                  }
                  break;
               case KEY_COMMENT_NAME :
               case KEY_COMMENT_DATE :
               case KEY_COMMENT_CREATOR :
	       case KEY_COMMENT_PROBLEM :
                  //(void)printf("Problem: [%s]\n", para[0].s);
		  if( strcmp(para[0].s, "Maximum Node Weight Connected Subgraph") == 0 )
		  {
		     stp_type = STP_MAX_NODE_WEIGHT;
		     //printf("Maximum Node Weight Connect \n");
		  }
                  break;
               case KEY_COMMENT_REMARK :
                  //(void)printf("Comment: [%s]\n", para[0].s);
                  break;
               case KEY_GRAPH_NODES :
                  nodes = (int)para[0].n;
                  break;
	       case KEY_GRAPH_ROOT :
		  stp_type = STP_REVENUES_BUDGET_HOPCONS;
		  root = (int)para[0].n - 1;
		  break;
	       case KEY_GRAPH_BUDGET :
		  stp_type = STP_REVENUES_BUDGET_HOPCONS;
		  budget = (int)para[0].n;
		  break;
	       case KEY_GRAPH_OBSTACLES :
		  nobstacles = (int)para[0].n;
		  if( nobstacles > 0 )
		     stp_type = STP_OBSTACLES_GRID;
                  break;
	       case KEY_GRAPH_HOPLIMIT :
		  //printf("HOP PROBLEM \n");

		  hoplimit = (int)para[0].n;
		  if( stp_type == UNKNOWN )
		     stp_type = STP_HOP_CONS;
		  break;
               case KEY_GRAPH_EDGES :
                  edges = (int)para[0].n;
                  break;
               case KEY_GRAPH_A :
               case KEY_GRAPH_E :
		  if( (int)para[0].n > nodes || (int)para[1].n > nodes )
                  {
                     message(MSG_FATAL, &curf, err_badedge_ddd,
                        (int)para[0].n, (int)para[1].n, nodes);
                     ret = FAILURE;
		     break;
                  }

                  if( g == NULL )
                  {
                     g = graph_init(nodes, edges * 2, 1, 0);

                     assert(g != NULL);
                     assert(g->source ==  UNKNOWN);
                     for( i = 0; i < nodes; i++ )
                        graph_knot_add(g, -1, 0, 0);

		     if( stp_type == STP_REVENUES_BUDGET_HOPCONS )
		     {
		        g->budget = budget;
			g->source = root;
			g->hoplimit = hoplimit;
		        g->stp_type = STP_REVENUES_BUDGET_HOPCONS;
		     }
                  }
                  if( stp_type == STP_HOP_CONS )
		  {
		     g->hoplimit = hoplimit;
		     tail = (int)para[0].n - 1;
		     head = (int)para[1].n - 1;
		     /* check whether the anti-parallel arc has already been added */
		     for( i = g->inpbeg[head]; i != EAT_LAST; i = g->ieat[i] )
		        if( g->tail[i] == tail )
			   break;
		     if( i == EAT_LAST )
		        graph_edge_add(g, tail, head, (double)para[2].n, FARAWAY);
		     else
		        g->cost[i] = (double)para[2].n;
		  }
                  else if( stp_type == STP_MAX_NODE_WEIGHT )
		  {
		     graph_edge_add(g, (int)para[0].n - 1, (int)para[1].n - 1, 0, 0);
		  }
                  else
                  {
                     graph_edge_add(g, (int)para[0].n - 1, (int)para[1].n - 1,
                        (double)para[2].n,
                        (p->sw_code == KEY_GRAPH_E)
                        ? (double)para[2].n
                        : (double)para[3].n);
                  }
                  break;
	       case KEY_MAXDEGS_MD :
                  //printf("MAX DEGS number  %d : ", degcount);
		  assert(g != NULL);
		  assert((int)para[0].n >= 0);

		  if( degcount < nodes )
		  {
                     if( g->maxdeg == NULL )
                     {
                        g->maxdeg = malloc((size_t)nodes * sizeof(int));
                        //printf("mode: DEG CONS \n");
			stp_type = STP_DEG_CONS;
			g->stp_type = STP_DEG_CONS;
                     }
                     g->maxdeg[degcount++] = (int)para[0].n;

                     //printf(", deg:  %d : \n", g->maxdeg[degcount - 1]);
                     break;
		  }
		  else
		  {
                     message(MSG_FATAL, &curf, err_baddeg_dd,
                        degcount, nodes);
                     ret = FAILURE;
		  }
	       case KEY_NODEWEIGHTS_NW :
		  nodeweight = (double) para[0].n;
		  if( g->knotcost == NULL )
		     g->knotcost = malloc((size_t)nodes * sizeof(int));

		  g->knotcost[nwcount] = nodeweight;
		  assert(g != NULL);
		  assert(presol != NULL);

		  if( stp_type != STP_NODE_WEIGHTS )
		     stp_type = STP_NODE_WEIGHTS;

		  if( Is_term(g->term[nwcount]) )
                     presol->fixed += nodeweight;
		  else
                     /* add node-weight to edge-weight of all incoming edges */
                     for( i = g->inpbeg[nwcount]; i != EAT_LAST; i = g->ieat[i] )
                        g->cost[i] += nodeweight;
		  nwcount++;
		  break;

	       case KEY_OBSTACLES_RR :
		  /*assert(nobstacles > 0);
                    if( obstacle_coords == NULL )
                    {
                    assert(obstacle_counter == 0);
                    obstacle_coords = (int**) malloc((size_t)4 * sizeof(int*));
                    for( i = 0; i < 4; i++ )
                    obstacle_coords[i] = (int*)malloc((size_t)nobstacles * sizeof(int));

                    }
                    for( i = 0; i < 4; i++ )
                    obstacle_coords[i][obstacle_counter] = (int)para[i].n;
                    obstacle_counter++;*/
		  break;
	       case KEY_OBSTACLES_END :
#if 0
		  message(MSG_INFO, &curf, "CRC [%X]", crc);
                  curf.section = &section_table[0];

		  if( obstacle_counter != nobstacles )
		  {
		     message(MSG_FATAL, &curf, "obstacle number does not match coordinates \n");
                     ret = FAILURE;
		     break;
		  }
                  if( scaled_coordinates == NULL )
		  {
		     message(MSG_FATAL, &curf, "coordinates not given \n");
                     ret = FAILURE;
		     break;
		  }
                  /* int f;
                     for( f = 0; f < nobstacles; f++ )
                     {
                     for( i = 0; i < 4; i++ )
		     printf("%d ", obstacle_coords[i][f]);
		     printf("\n");

                     }*/
                  assert(g == NULL);
                  g = graph_obstgrid_create(scaled_coordinates, obstacle_coords, nodes, grid_dim, nobstacles, scale_order);
                  for( i = 0; i < 4; i++ )
                     free(obstacle_coords[i]);
		  free(obstacle_coords);
#endif
		  break;

	       case KEY_PVERTICES_NPV :
		  g->npv = (int)para[0].n;
		  break;
	       case KEY_PVERTICES_PV :
		  if( g->knotcost == NULL )
		  {
		     int u;
                     g->knotcost = malloc((size_t)nodes * sizeof(double));
		     for( u = 0; u < nodes; u++ )
		        g->knotcost[u] = 0.0;
		  }
		  //printf("set : %d: %d \n", (int)para[0].n - 1, (int)para[1].n);
		  g->knotcost[(int)para[0].n - 1] = (int)para[1].n;
		  break;
	       case KEY_TERMINALS_END :
		  if( stp_type == STP_MAX_NODE_WEIGHT )
		  {
		     assert(nodes == termcount);
		     g->stp_type = STP_MAX_NODE_WEIGHT;
		     //graph_maxweight_transform(g, maxnodeweights);
		     //free(maxnodeweights);
		  }
                  else if( stp_type == STP_PRIZE_COLLECTING )
		  {
		     assert(prize != NULL);
		     g->stp_type = STP_PRIZE_COLLECTING;
                     //graph_prize_transform(g, prize);
		     //free(prize);
		  }
                  else if( stp_type == STP_ROOTED_PRIZE_COLLECTING )
		  {
		     assert(prize != NULL);
		     g->stp_type = STP_ROOTED_PRIZE_COLLECTING;
                     //graph_rootprize_transform(g, prize);
		     //free(prize);
		  }
		  curf.section = &section_table[0];
                  break;
               case KEY_TERMINALS_TERMINALS :
		  terms = (int)para[0].n;
		  assert(terms > 0);

		  if( stp_type == STP_MAX_NODE_WEIGHT )
		  {
                     assert(maxnodeweights == NULL);
                     assert(terms == nodes);
		     g->knotcost = malloc((size_t)nodes * sizeof(double));

                     maxnodeweights = g->knotcost;

		  }
                  break;
               case KEY_TERMINALS_ROOT :
                  assert(g != NULL);

                  if ((int)para[0].n <= nodes)
		  {
                     g->source = (int)para[0].n - 1;
		     graph_knot_chg(g, (int)para[0].n - 1, 0, NO_CHANGE, NO_CHANGE);
		  }
                  else
                  {
                     message(MSG_FATAL, &curf, err_badroot_dd,
                        (int)para[0].n, nodes);
                     ret = FAILURE;
                  }
                  break;
	       case KEY_TERMINALS_ROOTP :
		  assert(g != NULL);
		  assert(terms > 0);
		  g->source = (int)para[0].n - 1;
		  graph_knot_chg(g, (int)para[0].n - 1, 0, NO_CHANGE, NO_CHANGE);
		  assert(prize == NULL);
		  stp_type = STP_ROOTED_PRIZE_COLLECTING;
		  g->knotcost = malloc((size_t)nodes * sizeof(double));
		  prize = g->knotcost;
		  prize[(int)para[0].n - 1] = 0;
                  break;
               case KEY_TERMINALS_T :
		  if( stp_type == STP_MAX_NODE_WEIGHT )
		  {
		     assert(maxnodeweights != NULL);

		     maxnodeweights[(int)para[0].n - 1] = (double)para[1].n;
                     /* if( GT((double)para[1].n, 0.0) )
                        presol->fixed -= (double)para[1].n;*/

		     termcount++;
		  }
		  else
                     graph_knot_chg(g, (int)para[0].n - 1, 0, NO_CHANGE, NO_CHANGE);


                  break;
	       case KEY_TERMINALS_TP :
                  graph_knot_chg(g, (int)para[0].n - 1, 0, NO_CHANGE, NO_CHANGE);
		  if( stp_type != STP_PRIZE_COLLECTING && stp_type != STP_ROOTED_PRIZE_COLLECTING )
		  {
		     assert(g->knotcost == NULL);
		     stp_type = STP_PRIZE_COLLECTING;
		     g->knotcost = malloc((size_t)nodes * sizeof(double));
		     prize = g->knotcost;
		  }
		  prize[(int)para[0].n - 1] = (double)para[1].n;
		  //printf("prize[%d]: %f\n", (int)para[0].n , (double)para[1].n);
		  termcount++;
                  break;
               case KEY_COORDINATES_DD :
		  /* in this case coordinates are not needed */
		  if( terms > 0 )
		  {
		     ret        = SUCCESS;
                     stop_input = TRUE;
		     break;
		  }
		  //init_coordinates(g, para, &coordinates, &grid_dim, &termcount, 2, nodes);
                  break;
               case KEY_COORDINATES_DDD :
                  //init_coordinates(g, para, &coordinates, &grid_dim, &termcount, 3, nodes);
                  break;
	       case KEY_COORDINATES_DDDD :
		  //init_coordinates(g, para, &coordinates, &grid_dim, &termcount, 4, nodes);
                  break;
	       case KEY_COORDINATES_DDDDD :
		  //init_coordinates(g, para, &coordinates, &grid_dim, &termcount, 5, nodes);
                  break;
	       case KEY_COORDINATES_DDDDDD :
		  //init_coordinates(g, para, &coordinates, &grid_dim, &termcount, 6, nodes);
                  break;
	       case KEY_COORDINATES_DDDDDDD :
		  //init_coordinates(g, para, &coordinates, &grid_dim, &termcount, 7, nodes);
                  break;
	       case KEY_COORDINATES_DDDDDDDD :
		  //init_coordinates(g, para, &coordinates, &grid_dim, &termcount, 8, nodes);
                  break;
	       case KEY_COORDINATES_END :
#if 0
		  assert(g == NULL);
		  assert(grid_dim > 1);
		  message(MSG_INFO, &curf, "CRC [%X]", crc);

                  curf.section = &section_table[0];
		  if( termcount != nodes )
		  {
		     message(MSG_FATAL, &curf, "node number does not match coordinates \n");
                     ret = FAILURE;
		     break;
		  }

		  /* scale all coordinates such that they are integers */
		  scale_coords(coordinates, &scaled_coordinates, &scale_order, nodes, grid_dim);

		  for( i = 0; i < grid_dim; i++ )
                     free(coordinates[i]);

		  free(coordinates);

		  if( stp_type != STP_OBSTACLES_GRID )
		     g = graph_grid_create(scaled_coordinates, nodes, grid_dim, scale_order);
#endif
		  break;
               case KEY_COORDINATES_GRID :
                  is_gridgraph = TRUE;
                  break;
               case KEY_PRESOLVE_FIXED :
                  if (presol != NULL)
                     presol->fixed += (double)para[0].n;
                  break;
               case KEY_PRESOLVE_DATE :
                  //(void)printf("Found presolve information %s\n",
                  //para[0].s);
                  break;
               case KEY_PRESOLVE_LOWER :
                  if (presol != NULL)
                     presol->lower = (double)para[0].n;
                  break;
               case KEY_PRESOLVE_UPPER :
                  if (presol != NULL)
                     presol->upper = (double)para[0].n;
                  break;
               case KEY_PRESOLVE_TIME :
                  if (presol != NULL)
                     presol->time = (int)para[0].n;
                  break;
	       case KEY_PRESOLVE_EA :
		  break;
	       case KEY_PRESOLVE_EC :
		  break;
	       case KEY_PRESOLVE_ED :
		  break;
	       case KEY_PRESOLVE_ES :
		  break;
               default :
                  /* CONSTCOND */
                  assert(FALSE);
                  break;
               }
            }
         }
      }
   }
   /* Was there an error in an incuded file ?
    * Than close the main file.
    */
   if (save.fp != NULL)
      (void)fclose(save.fp);

   /* Close the actual file anyway. Since we stop at encountering
    * a line with "EOF" on it, this will be the normal case.
    */
   if (curf.fp != NULL)
      (void)fclose(curf.fp);

   if (ret == SUCCESS)
   {
      assert(g != NULL);

      if( g->source == UNKNOWN )
      {
         for(i = 0; i < g->knots; i++)
            if ((g->term[i] == 0)
               && ((g->source < 0) || (g->grad[i] > g->grad[g->source])))
               g->source = i;
      }

      if( g->stp_type == UNKNOWN )
      {
	 if( stp_type != UNKNOWN )
	    g->stp_type = stp_type;
	 else
            g->stp_type = STP_UNDIRECTED;
      }

      //(void)printf(msg_finish_dddd,
      //g->knots, g->edges, g->terms, g->source);
      if( 0 && g->stp_type != STP_MAX_NODE_WEIGHT && g->stp_type != STP_REVENUES_BUDGET_HOPCONS )
      {
         assert(graph_valid(g));
      }
   }
   return((ret == SUCCESS) ? g : NULL);
}



/*
 * Minimum Spanning Tree methods
 */

static int mst_heap_nearest(
   int* heap,
   int* state,
   int* count,    /* pointer to store the number of elements on the heap */
   const MST* path)
{
   int   k;
   int   t;
   int   c;
   int   j;

   /* Heap shift down
    */
   k              = heap[1];
   j              = 1;
   c              = 2;
   heap[1]        = heap[(*count)--];
   state[heap[1]] = 1;

   if ((*count) > 2)
      if (LT(path[heap[3]].dist, path[heap[2]].dist))
         c++;

   while((c <= (*count)) && GT(path[heap[j]].dist, path[heap[c]].dist))
   {
      t              = heap[c];
      heap[c]        = heap[j];
      heap[j]        = t;
      state[heap[j]] = j;
      state[heap[c]] = c;
      j              = c;
      c             += c;

      if ((c + 1) <= (*count))
         if (LT(path[heap[c + 1]].dist, path[heap[c]].dist))
            c++;
   }
   return(k);
}

static void mst_heap_correct(
   int* heap,
   int* state,
   int* count,    /* pointer to store the number of elements on the heap */
   MST*  path,
   int    l,
   int    k,
   int    e,
   double cost)
{
   int    t;
   int    c;
   int    j;

   path[l].dist = cost;
   path[l].edge = e;

   /* Has the node not been processed yet?
    */
   if (state[l] == UNKNOWN)
   {
      heap[++(*count)] = l;
      state[l]      = (*count);
   }

   /* Heap shift up
    */
   j = state[l];
   c = j / 2;

   while((j > 1) && GT(path[heap[c]].dist, path[heap[j]].dist))
   {
      t              = heap[c];
      heap[c]        = heap[j];
      heap[j]        = t;
      state[heap[j]] = j;
      state[heap[c]] = c;
      j              = c;
      c              = j / 2;
   }
}

/** initialize a Minimum Spanning Tree on the graph */
static void mst_init(
   GRAPH* p)
{
   assert(p->path_heap  == NULL);
   assert(p->path_state == NULL);

   p->path_heap  = (int*)malloc((size_t)p->knots * sizeof(int));
   p->path_state = (int*)malloc((size_t)p->knots * sizeof(int));

   assert(p->path_heap  != NULL);
   assert(p->path_state != NULL);
}

/** free the Minimum Spanning Tree data structures on the graph */
static void mst_exit(
   GRAPH* p)
{
   assert(p->path_heap  != NULL);
   assert(p->path_state != NULL);

   free(p->path_heap);
   free(p->path_state);

   p->path_heap  = NULL;
   p->path_state = NULL;
}

/** execute a Minimum Spanning Tree computation */
static void mst_exec(
   const GRAPH*  p,
   int           start,
   double*       cost,
   MST*          path)
{
   int   k;
   int   m;
   int   i;
   int* heap;
   int* state;
   int count;

   assert(p      != NULL);
   assert(start  >= 0);
   assert(start  <  p->knots);
   assert(p->path_heap   != NULL);
   assert(p->path_state  != NULL);
   assert(path   != NULL);
   assert(cost   != NULL);

   /* if there are no nodes, terminate
    */
   if (p->knots == 0)
      return;

   heap = p->path_heap;
   state = p->path_state;

   /* initalize all needed data structures
    */
   for(i = 0; i < p->knots; i++)
   {
      state[i]     = UNKNOWN;
      path[i].dist = FARAWAY;
      path[i].edge = -1;
   }
   /* add the starting node to the heap
    */
   k            = start;
   path[k].dist = 0.0;

   /* at least two nodes are necessary for the heap to work
    */
   if (p->knots > 1)
   {
      count       = 1;
      heap[count] = k;
      state[k]    = count;

      /* as soon as the heap is empty the computation has been finished
       */
      while( count > 0 )
      {
         /* get a nearest node
          */
         k = mst_heap_nearest(heap, state, &count, path);

         /* mark the node as scanned
          */
         state[k] = CONNECT;

         /* correct adjacent nodes
          */
         for(i = p->outbeg[k]; i != EAT_LAST; i = p->oeat[i])
         {
            m = p->head[i];

            /* has the node not been scanned yet?
             * is it forbidded?
             */
            if ((state[m]) && (p->mark[m])

               /* is the new connection shorter than the currently best known?
                */
               && (GT(path[m].dist, cost[i])))
               mst_heap_correct(heap, state, &count, path, m, k, i, cost[i]);
         }
      }
   }
}


static void trim_right(
   char* s)
{
   size_t i;
   i = 0;
   while(s[i] != '\0' && isspace((unsigned char)s[i]))
      i++;
   memmove(s, s + i, strlen(s) - i + 1);
}

static void trim_left(
   char* s)
{
   size_t i;
   i = strlen(s);
   while(i > 0 && isspace((unsigned char)s[i - 1]))
      i--;
   s[i] = '\0';
}

/* delete whitespace both from the left- and the right-hand side */
static void trim(
   char* s)
{
   trim_right(s);
   trim_left(s);
}



/*
 * Clock
 */
static double clock_cputime2sec(
   clock_t cputime)             /* clock ticks for CPU time */
{
   clock_t clocks_per_second;

#if defined(_WIN32) || defined(_WIN64)
   clocks_per_second = 100;
#else
#ifndef CLK_TCK
   clocks_per_second = sysconf(_SC_CLK_TCK);
#else
   clocks_per_second = CLK_TCK;
#endif
#endif

   return (double)(cputime) / (double)clocks_per_second;
}


static void clock_start(
   clock_t* cputime)
{
   assert(cputime != NULL);

   *cputime = 0;
#if defined(_WIN32) || defined(_WIN64)
   FILETIME creationtime;
   FILETIME exittime;
   FILETIME kerneltime;
   FILETIME usertime;

   GetProcessTimes(GetCurrentProcess(), &creationtime, &exittime, &kerneltime, &usertime);
   *cputime -= usertime.dwHighDateTime * 42950 + usertime.dwLowDateTime / 100000L;
#else
   struct timeval tp;
   struct tms now;
   (void)times(&now);
   *cputime -= now.tms_utime;
#endif

}


static double clock_stop(
   clock_t* cputime)
{
   double result;
   assert(cputime != NULL);
   result = 0.0;
#if defined(_WIN32) || defined(_WIN64)
   FILETIME creationtime;
   FILETIME exittime;
   FILETIME kerneltime;
   FILETIME usertime;

   GetProcessTimes(GetCurrentProcess(), &creationtime, &exittime, &kerneltime, &usertime);
   result = clock_cputime2sec(*cputime + usertime.dwHighDateTime * 42950 + usertime.dwLowDateTime / 100000L);
#else
   struct timeval tp;
   struct tms now;

   (void)times(&now);
   result = clock_cputime2sec(*cputime + now.tms_utime);
#endif

   return result;
}


/*
 * Main
 */
int main(
   int argc,
   char **argv)
{
   GRAPH*     graph;
   MST*       mst;
   FILE*      probfp;
   FILE*      solfp;
   double     solval;
   double     soltime;
   int        e;
   int        i;
   int        j;
   int        start;
   int        edgecount = 0;
   int        v = -1;
   double     cost = 0;
   double     revenue = 0.0;
   int        nsolnodes = -1;
   int        nsoledges = -1;
   int        root;
   int        nnodes;
   int        nedges;
   int        deg;
   char*      edges;
   char*      nodes;
   char*      probfilename;
   char*      solfilename;
   char*      s;
   char*      line;
   char       mode[128];
   char       solsection = FALSE;
   char       graphsection = FALSE;
   char       buffer[MAX_LINE_LEN];

   if( argc != 3 )
   {
      printf("Wrong number of arguments (should be two) \n");
      return 0;
   }
   probfilename = argv[1];
   solfilename = argv[2];

   trim(probfilename);
   trim(solfilename);

   /* read the solution file */
   if( (solfp = fopen(solfilename, "r")) == NULL )
   {
      printf("Solution file missing \n");
      return 0;
   }

   /* load the graph */
   graph = graph_load(probfilename, NULL);
   if( graph == NULL )
   {
      printf("Graph could not be read \n");
      return 0;
   }

   nnodes = graph->knots;
   nedges = graph->edges;

   nodes = malloc((size_t)nnodes * sizeof(char));
   edges = malloc((size_t)nedges * sizeof(char));

   for( i = 0; i < nnodes; i++ )
   {
      nodes[i] = FALSE;
      graph->mark[i] = TRUE;
   }

   for( i = 0; i < nedges; i++ )
      edges[i] = FALSE;

   /* read the solution, line per line */
   while( solfp != NULL )
   {
      if( (line = fgets(buffer, sizeof(buffer), solfp)) == NULL )
      {
         (void)fclose(solfp);
         break;
      }
      /* delete trailing whitespace */
      trim(line);

      if( strcmp(solfilename, "EOF") == 0 )
      {
	 (void)fclose(solfp);
         break;
      }
      if (*line == '\0')
         continue;

      if( !solsection )
      {
         if( strcmp(strlower(line), "section run") == 0 )
         {
            solsection = TRUE;
	    continue;
         }
      }
      else
      {
	 sscanf(line, "%s", mode);

	 if( strcmp(strlower(line), "end") == 0 )
	 {
	    solsection = FALSE;
	    continue;
	 }
	 else if( strcmp(strlower(mode), "primal") == 0 )
	 {
	    sscanf(line, "%s %lf" , mode,  &solval);
	 }
      }

      if( !graphsection )
      {
         if( strcmp(strlower(line), "section finalsolution") == 0 )
         {
            graphsection = TRUE;
         }
         continue;
      }

      sscanf(line, "%s %d %d", mode, &i, &j);

      if( strcmp(strlower(strlower(mode)), "vertices") == 0 )
      {
	 nsolnodes = i;
	 assert(nsolnodes >= 0);
      }
      else if( strcmp(strlower(mode), "v") == 0 )
      {
	 assert(i >= 1);
         nodes[i - 1] = TRUE;
	 v = i - 1;
      }
      else if( strcmp(strlower(mode), "edges") == 0 )
      {
	 nsoledges = i;
	 assert(nsoledges >= 0);
      }
      else if( strcmp(strlower(mode), "e") == 0 )
      {
	 assert(i >= 1);
	 assert(j >= 1);
	 for( e = graph->outbeg[i - 1]; e != EAT_LAST ; e = graph->oeat[e] )
         {
            if( graph->head[e] == j - 1 )
	    {
	       cost += graph->cost[e];
	       edges[e] = TRUE;
	       edges[Edge_anti(e)] = TRUE;
	       edgecount++;
	       break;
	    }
         }
      }
      else if( strcmp(strlower(mode), "end") == 0 )
      {
	 (void)fclose(solfp);
	 break;
      }
   }

   for( i = 0; i < nedges; i += 2 )
   {
      if( edges[i] == FALSE )
         graph_edge_del(graph, i);
   }

   if( v == -1 && nsolnodes != 0 )
   {
      printf("solution has no vertices (%d reported) \n", nsolnodes);
      return 0;
   }
   
   /* allow empty tree for PCSTP and MWCS */
   if( v == -1 )
   {
      if( graph->stp_type != STP_PRIZE_COLLECTING && graph->stp_type != STP_MAX_NODE_WEIGHT )
      {
         printf("solution has no vertices \n");
         return 0;
      }
      else
      {
         for( i = 0; i < nnodes; i++ )
            cost += graph->knotcost[i];
         if( NE(cost, solval) )
         {
            printf("###sol 0\n");
            printf("Real costs: %f; reported costs: %f\n", cost, solval);
            return 0;
         }
      }
   }

   /* initialize a MST on the current graph */
   mst = (MST*)malloc((size_t) nnodes * sizeof(*mst));
   graph->path_state = NULL;
   mst_init(graph);

   if( graph->stp_type == STP_UNDIRECTED || graph->stp_type == STP_DEG_CONS || graph->stp_type == STP_ROOTED_PRIZE_COLLECTING
      || graph->stp_type == STP_HOP_CONS || graph->stp_type == STP_REVENUES_BUDGET_HOPCONS || graph->stp_type == STP_NODE_WEIGHTS )
   {
      start = graph->source;
   }
   else if( graph->stp_type == STP_PRIZE_COLLECTING || graph->stp_type == STP_MAX_NODE_WEIGHT )
   {
      start = v;
   }
   else
   {
      printf("unknow problem type \n");
      return 0;
   }

   mst_exec(graph, start, graph->cost, mst);

   if( graph->stp_type == STP_PRIZE_COLLECTING || graph->stp_type == STP_ROOTED_PRIZE_COLLECTING )
   {
      for( i = 0; i < nnodes; i++ )
      {
         if( Is_term(graph->term[i]) && graph->path_state[i] != CONNECT )
	 {
            cost += graph->knotcost[i];
	 }
      }
   }

   if( graph->stp_type == STP_MAX_NODE_WEIGHT || graph->stp_type == STP_REVENUES_BUDGET_HOPCONS ||
      graph->stp_type == STP_NODE_WEIGHTS )
   {
      for( i = 0; i < nnodes; i++ )
      {
         if( graph->path_state[i] == CONNECT )
	 {
	    if( graph->stp_type == STP_REVENUES_BUDGET_HOPCONS )
	       revenue += graph->knotcost[i];
	    else
               cost += graph->knotcost[i];
	 }
      }
   }

   if( graph->stp_type == STP_HOP_CONS && graph->hoplimit < edgecount )
   {
      printf("###sol 0\n");
      printf("Hop limit violated: %d < %d \n", graph->hoplimit, edgecount );
      return 0;
   }

   if( graph->stp_type == STP_UNDIRECTED || graph->stp_type == STP_DEG_CONS || graph->stp_type == STP_HOP_CONS
      || graph->stp_type == STP_REVENUES_BUDGET_HOPCONS )
   {
      for( i = 0; i < nnodes; i++ )
      {
         if( nodes[i] == TRUE )
         {
            if( graph->path_state[i] != CONNECT )
            {
               printf("###sol 0\n");
	       printf("Vertex %d not connected \n", i+ 1);
               return 0;
            }
            if( graph->stp_type == STP_DEG_CONS )
            {
               deg = 0;
               for( e = graph->outbeg[i]; e != EAT_LAST ; e = graph->oeat[e] )
                  if( edges[e] == TRUE )
                     deg++;
               if( deg > graph->maxdeg[i] )
               {
                  printf("###sol 0\n");
		  printf("Degree constraint for vertex %d violated \n", i + 1);
                  return 0;
               }
            }
         }
      }
   }

   if( graph->stp_type != STP_REVENUES_BUDGET_HOPCONS && NE(cost, solval) )
   {
      printf("###sol 0\n");
      printf("Real costs: %f; reported costs: %f\n", cost, solval);
      return 0;
   }
   else if( graph->stp_type == STP_REVENUES_BUDGET_HOPCONS )
   {
      // printf("Costs: %f; budget: %d\n", cost, graph->budget);
      // printf("Real revenue: %f; reported revenue: %f\n", revenue, solval);
      if( NE(revenue, solval) )
      {
         printf("###sol 0\n");
         printf("Real revenue: %f; reported revenue: %f\n", revenue, solval);
         return 0;
      }
      else if( GT(cost, graph->budget) )
      {
	 printf("###sol 0\n");
         printf("Costs: %f; budget: %d\n", cost, graph->budget);
         return 0;
      }
      /* test hop constraints */
      for( i = 0; i < nnodes; i++ )
      {
         if( nodes[i] == TRUE )
	 {
            v = i;
            j = 0;
            while( v != graph->source )
            {
               v = graph->tail[mst[v].edge];
               j++;
               //       printf("++ ");
            }
            //    printf("\n");
            if( j > graph->hoplimit )
            {
               printf("###sol 0\n");
               printf("Hop limit violated for terminal %d\n", i + 1);
               return 0;
            }
	 }
      }
   }
   printf("###sol 1\n");

   /* free memory */
   mst_exit(graph);
   free(mst);
   free(nodes);
   free(edges);
   graph_free(graph);
   return 1;
}
